/*
** based on
**
** File: ymf262.c - software implementation of YMF262
**                  FM sound generator type OPL3
**
** Copyright (C) 2003 Jarek Burczynski
**
** Version 0.2
**
** from the dosbox 0.72 source
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "ymf262.h"
#include "940/ymf262_shared.h"

#ifndef PI
#define PI 3.14159265358979323846
#endif

/* output final shift */
#if (OPL3_SAMPLE_BITS==16)
	#define FINAL_SH	(0)
	#define MAXOUT		(+32767)
	#define MINOUT		(-32768)
#else
	#define FINAL_SH	(8)
	#define MAXOUT		(+127)
	#define MINOUT		(-128)
#endif


#define FREQ_SH			16  /* 16.16 fixed point (frequency calculations) */
#define EG_SH			16  /* 16.16 fixed point (EG timing)              */
#define LFO_SH			24  /*  8.24 fixed point (LFO calculations)       */
#define TIMER_SH		16  /* 16.16 fixed point (timers calculations)    */

#define FREQ_MASK		((1<<FREQ_SH)-1)

/* envelope output entries */
#define ENV_BITS		10
#define ENV_LEN			(1<<ENV_BITS)
#define ENV_STEP		(128.0/ENV_LEN)

#define MAX_ATT_INDEX	((1<<(ENV_BITS-1))-1) /*511*/
#define MIN_ATT_INDEX	(0)

/* register number to channel number , slot offset */
#define SLOT1 0
#define SLOT2 1

/* Envelope Generator phases */

#define EG_ATT			4
#define EG_DEC			3
#define EG_SUS			2
#define EG_REL			1
#define EG_OFF			0

#define OPL3_TYPE_YMF262 (0)	/* 36 operators, 8 waveforms */

#define WAIT_IF_MSG_BUF_FULL while( *NSubmittedMessages - *NExecutedMessages>= MSG_BUF_SIZE){};

#define WAIT_FOR_SYNC \
 while( (*NSubmittedMessages - *NExecutedMessages) % (int) MSG_BUF_SIZE !=0 ){};

static int counter[8];

#define ADD_MESSAGE(mtype, i,j,k) \
 { \
 int n = ((*NSubmittedMessages)+1) % ((int) MSG_BUF_SIZE); \
 MessageBuffer[n].type=mtype;  \
 MessageBuffer[n].data1=i;  \
 MessageBuffer[n].data2=j;  \
 MessageBuffer[n].data3=k; \
 (*NSubmittedMessages)++;  \
 }

/* OPL3 state */
typedef struct {
	UINT8	OPL3_mode;				/* OPL3 extension enable flag	*/
	int	T[2];					/* timer counters				*/
	int	TC[2];
	UINT8	st[2];					/* timer enable					*/

	UINT32  *fn_tab;

	UINT32	address;				/* address register				*/
	UINT8	status;					/* status flag					*/
	UINT8	statusmask;				/* status mask					*/

	/* external event callback handlers */
	OPL3_TIMERHANDLER  TimerHandler;/* TIMER handler				*/
	int TimerParam;					/* TIMER parameter				*/
	OPL3_IRQHANDLER    IRQHandler;	/* IRQ handler					*/
	int IRQParam;					/* IRQ parameter				*/
	OPL3_UPDATEHANDLER UpdateHandler;/* stream update handler		*/
	int UpdateParam;				/* stream update parameter		*/

	UINT8 	type;						/* chip type					*/
	int 	clock;						/* master clock  (Hz)			*/
	int 	rate;						/* sampling rate (Hz)			*/
	double 	freqbase;				/* frequency base				*/
	double 	TimerBase;				/* Timer base time (==sampling time)*/
} OPL3lite;

/* status set and IRQ handling */
INLINE void OPL3_STATUS_SET(OPL3lite *chip,int flag)
{
	/* set status flag masking out disabled IRQs */
	chip->status |= (flag & chip->statusmask);
	if(!(chip->status & 0x80))
	{
		if(chip->status & 0x7f)
		{	/* IRQ on */
			chip->status |= 0x80;
			/* callback user interrupt handler (IRQ is OFF to ON) */
			if(chip->IRQHandler) (chip->IRQHandler)(chip->IRQParam,1);
		}
	}
}

/* status reset and IRQ handling */
INLINE void OPL3_STATUS_RESET(OPL3lite *chip,int flag)
{
	/* reset status flag */
	chip->status &= ~flag;
	if(chip->status & 0x80)
	{
		if (!(chip->status & 0x7f))
		{
			chip->status &= 0x7f;
			/* callback user interrupt handler (IRQ is ON to OFF) */
			if(chip->IRQHandler) (chip->IRQHandler)(chip->IRQParam,0);
		}
	}
}

/* IRQ mask set */
INLINE void OPL3_STATUSMASK_SET(OPL3lite *chip,int flag)
{
	chip->statusmask = flag;
	/* IRQ handling check */
	OPL3_STATUS_SET(chip,0);
	OPL3_STATUS_RESET(chip,0);
}

/* generic table initialize */
static int init_tables(void)
{
	signed int i,x;
	signed int n;
	double o,m;


	for (x=0; x<TL_RES_LEN; x++)
	{
		m = (1<<16) / pow(2.0, (x+1) * (ENV_STEP/4.0) / 8.0);
		m = floor(m);

		/* we never reach (1<<16) here due to the (x+1) */
		/* result fits within 16 bits at maximum */

		n = (int)m;		/* 16 bits here */
		n >>= 4;		/* 12 bits here */
		if (n&1)		/* round to nearest */
			n = (n>>1)+1;
		else
			n = n>>1;
						/* 11 bits here (rounded) */
		n <<= 1;		/* 12 bits here (as in real chip) */
		tl_tab[ x*2 + 0 ] = n;
		tl_tab[ x*2 + 1 ] = ~tl_tab[ x*2 + 0 ]; /* this *is* different from OPL2 (verified on real YMF262) */

		for (i=1; i<13; i++)
		{
			tl_tab[ x*2+0 + i*2*TL_RES_LEN ] =  tl_tab[ x*2+0 ]>>i;
			tl_tab[ x*2+1 + i*2*TL_RES_LEN ] = ~tl_tab[ x*2+0 + i*2*TL_RES_LEN ];  /* this *is* different from OPL2 (verified on real YMF262) */
		}
	}

	for (i=0; i<SIN_LEN; i++)
	{
		/* non-standard sinus */
		m = sin( ((i*2)+1) * PI / SIN_LEN ); /* checked against the real chip */

		/* we never reach zero here due to ((i*2)+1) */

		if (m>0.0)
			o = 8*log(1.0/m)/log(2.0);	/* convert to 'decibels' */
		else
			o = 8*log(-1.0/m)/log(2.0);	/* convert to 'decibels' */

		o = o / (ENV_STEP/4);

		n = (int)(2.0*o);
		if (n&1)						/* round to nearest */
			n = (n>>1)+1;
		else
			n = n>>1;

		sin_tab[ i ] = n*2 + (m>=0.0? 0: 1 );

		/*logerror("YMF262.C: sin [%4i (hex=%03x)]= %4i (tl_tab value=%5i)\n", i, i, sin_tab[i], tl_tab[sin_tab[i]] );*/
	}

	for (i=0; i<SIN_LEN; i++)
	{
		/* these 'pictures' represent _two_ cycles */
		/* waveform 1:  __      __     */
		/*             /  \____/  \____*/
		/* output only first half of the sinus waveform (positive one) */

		if (i & (1<<(SIN_BITS-1)) )
			sin_tab[1*SIN_LEN+i] = TL_TAB_LEN;
		else
			sin_tab[1*SIN_LEN+i] = sin_tab[i];

		/* waveform 2:  __  __  __  __ */
		/*             /  \/  \/  \/  \*/
		/* abs(sin) */

		sin_tab[2*SIN_LEN+i] = sin_tab[i & (SIN_MASK>>1) ];

		/* waveform 3:  _   _   _   _  */
		/*             / |_/ |_/ |_/ |_*/
		/* abs(output only first quarter of the sinus waveform) */

		if (i & (1<<(SIN_BITS-2)) )
			sin_tab[3*SIN_LEN+i] = TL_TAB_LEN;
		else
			sin_tab[3*SIN_LEN+i] = sin_tab[i & (SIN_MASK>>2)];

		/* waveform 4:                 */
		/*             /\  ____/\  ____*/
		/*               \/      \/    */
		/* output whole sinus waveform in half the cycle(step=2) and output 0 on the other half of cycle */

		if (i & (1<<(SIN_BITS-1)) )
			sin_tab[4*SIN_LEN+i] = TL_TAB_LEN;
		else
			sin_tab[4*SIN_LEN+i] = sin_tab[i*2];

		/* waveform 5:                 */
		/*             /\/\____/\/\____*/
		/*                             */
		/* output abs(whole sinus) waveform in half the cycle(step=2) and output 0 on the other half of cycle */

		if (i & (1<<(SIN_BITS-1)) )
			sin_tab[5*SIN_LEN+i] = TL_TAB_LEN;
		else
			sin_tab[5*SIN_LEN+i] = sin_tab[(i*2) & (SIN_MASK>>1) ];

		/* waveform 6: ____    ____    */
		/*                             */
		/*                 ____    ____*/
		/* output maximum in half the cycle and output minimum on the other half of cycle */

		if (i & (1<<(SIN_BITS-1)) )
			sin_tab[6*SIN_LEN+i] = 1;	/* negative */
		else
			sin_tab[6*SIN_LEN+i] = 0;	/* positive */

		/* waveform 7:                 */
		/*             |\____  |\____  */
		/*                   \|      \|*/
		/* output sawtooth waveform    */

		if (i & (1<<(SIN_BITS-1)) )
			x = ((SIN_LEN-1)-i)*16 + 1;	/* negative: from 8177 to 1 */
		else
			x = i*16;	/*positive: from 0 to 8176 */

		if (x > TL_TAB_LEN)
			x = TL_TAB_LEN;	/* clip to the allowed range */

		sin_tab[7*SIN_LEN+i] = x;

	}

	return 1;
}


static void OPL3_initalize(OPL3lite *chip,OPL3 *chips)
{
	int i;

	/* frequency base */
	chip->freqbase  = (chip->rate) ? ((double)chip->clock / (8.0*36)) / chip->rate  : 0;

	/* Timer base time */
	chip->TimerBase = 1.0 / ((double)chip->clock / (8.0*36) );

	/* make fnumber -> increment counter table */

	for( i=0 ; i < 1024 ; i++ )
	{
		/* opn phase increment counter = 20bit */
		chip->fn_tab[i] = (UINT32)( (double) i * 64 * chip->freqbase * (1<<(FREQ_SH-10)) ); /* -10 because chip works with 10.10 fixed point, while we use 16.16 */
	}
// 	chips->fn_max=(long long int) ( (double) 1024 * 64 * chip->freqbase*(1<<(FREQ_SH-10)));
//        LOG_MSG("fnmax: %lld %lld %d",chips->fn_max,1024*chips->fn_max,sizeof(long long int));

	/* Amplitude modulation: 27 output levels (triangle waveform); 1 level takes one of: 192, 256 or 448 samples */
	/* One entry from LFO_AM_TABLE lasts for 64 samples */
	chips->lfo_am_inc = (UINT32)((1.0 / 64.0 ) * (1<<LFO_SH) * chip->freqbase);

	/* Vibrato: 8 output levels (triangle waveform); 1 level takes 1024 samples */
	chips->lfo_pm_inc = (UINT32)((1.0 / 1024.0) * (1<<LFO_SH) * chip->freqbase);

	/* Noise generator: a step takes 1 sample */
	chips->noise_f = (UINT32)((1.0 / 1.0) * (1<<FREQ_SH) * chip->freqbase);

	chips->eg_timer_add  = (UINT32)((1<<EG_SH)  * chip->freqbase);
	chips->eg_timer_overflow = ( 1 ) * (1<<EG_SH);

}

/* write a value v to register r on OPL chip */
static void OPL3WriteReg(OPL3lite *chip, int r, int v)
{
	if(r&0x100)
	{
		switch(r)
		{
		case 0x101:	/* test register */
			return;
		break;
		case 0x104:	/* 6 channels enable */
			return;
		break;
		case 0x105:	/* OPL3 extensions enable register */
			chip->OPL3_mode = v&0x01;	/* OPL3 mode when bit0=1 otherwise it is OPL2 mode */
			return;
		break;

		default:
			if (r < 0x120)
				logerror("YMF262: write to unknown register (set#2): %03x value=%02x\n",r,v);
		break;
		}
	}
	/* adjust bus to 8 bits */
	r &= 0xff;
	v &= 0xff;

	switch(r&0xe0)
	{
	case 0x00:	/* 00-1f:control */
		switch(r&0x1f)
		{
		case 0x01:	/* test register */
		break;
		case 0x02:	/* Timer 1 */
			chip->T[0] = (256-v)*4;
		break;
		case 0x03:	/* Timer 2 */
			chip->T[1] = (256-v)*16;
		break;
		case 0x04:	/* IRQ clear / mask and Timer enable */
			if(v&0x80)
			{	/* IRQ flags clear */
				OPL3_STATUS_RESET(chip,0x60);
			}
			else
			{	/* set IRQ mask ,timer enable */
				chip->st[0] = v & 1;
				chip->st[1] = (v>>1) & 1;

				/* IRQRST,T1MSK,t2MSK,x,x,x,ST2,ST1 */
				OPL3_STATUS_RESET(chip, v & 0x60);
				OPL3_STATUSMASK_SET(chip, (~v) & 0x60 );

				/* timer 1 */
				if(chip->st[0])
				{
					chip->TC[0]=chip->T[0]*20;
					double interval = (double)chip->T[0]*chip->TimerBase;
					if (chip->TimerHandler) (chip->TimerHandler)(chip->TimerParam+0,interval);
				}
				/* timer 2 */
				if(chip->st[1])
				{
					chip->TC[1]=chip->T[1]*20;
					double interval =(double)chip->T[1]*chip->TimerBase;
					if (chip->TimerHandler) (chip->TimerHandler)(chip->TimerParam+1,interval);
				}
			}
		break;
		case 0x08:	/* x,NTS,x,x, x,x,x,x */
		break;

		default:
			logerror("YMF262: write to unknown register: %02x value=%02x\n",r,v);
		break;
		}
		break;
	case 0x20:	/* am ON, vib ON, ksr, eg_type, mul */
	break;
	case 0x40:
	break;
	case 0x60:
	break;
	case 0x80:
	break;
	case 0xa0:
	break;
	case 0xc0:
	break;
	case 0xe0: /* waveform select */
	break;
	}
}


static void OPL3ResetChip(OPL3lite *chip)
{
	int c,s;

	OPL3_STATUS_RESET(chip,0x60);

	/* reset with register write */
	OPL3WriteReg(chip,0x01,0); /* test register */
	OPL3WriteReg(chip,0x02,0); /* Timer1 */
	OPL3WriteReg(chip,0x03,0); /* Timer2 */
	OPL3WriteReg(chip,0x04,0); /* IRQ mask clear */
//FIX IT  registers 101, 104 and 105
//FIX IT (dont change CH.D, CH.C, CH.B and CH.A in C0-C8 registers)
	for(c = 0xff ; c >= 0x20 ; c-- )
		OPL3WriteReg(chip,c,0);
//FIX IT (dont change CH.D, CH.C, CH.B and CH.A in C0-C8 registers)
	for(c = 0x1ff ; c >= 0x120 ; c-- )
		OPL3WriteReg(chip,c,0);
}

/* Create one of virtual YMF262 */
/* 'clock' is chip clock in Hz  */
/* 'rate'  is sampling rate  */
void OPL3Create(int type, int clock, int rate,OPL3lite *chip,OPL3 *chips)
{
	chip->type  = type;
	chip->clock = clock;
	chip->rate  = rate;

	/* init global tables */
	OPL3_initalize(chip,chips);
}

/* Destroy one of virtual YMF262 */
static void OPL3Destroy(OPL3lite *chip)
{
	free(chip);
}


/* Optional handlers */

static void OPL3SetTimerHandler(OPL3lite *chip,OPL3_TIMERHANDLER TimerHandler,int channelOffset)
{
	chip->TimerHandler   = TimerHandler;
	chip->TimerParam = channelOffset;
}
static void OPL3SetIRQHandler(OPL3lite *chip,OPL3_IRQHANDLER IRQHandler,int param)
{
	chip->IRQHandler     = IRQHandler;
	chip->IRQParam = param;
}
static void OPL3SetUpdateHandler(OPL3lite *chip,OPL3_UPDATEHANDLER UpdateHandler,int param)
{
	chip->UpdateHandler = UpdateHandler;
	chip->UpdateParam = param;
}

/* YMF262 I/O interface */
static int OPL3Write(OPL3lite *chip, int a, int v)
{
	/* data bus is 8 bits */
	v &= 0xff;

	switch(a&3)
	{
	case 0:	/* address port 0 (register set #1) */
		chip->address = v;
	break;

	case 1:	/* data port - ignore A1 */
	case 3:	/* data port - ignore A1 */
		if(chip->UpdateHandler) chip->UpdateHandler(chip->UpdateParam,0);
		OPL3WriteReg(chip,chip->address,v);
	break;

	case 2:	/* address port 1 (register set #2) */

		if( chip->OPL3_mode & 1 )
		{
			/* OPL3 mode */
				chip->address = v | 0x100;
		}
		else
		{
			/* in OPL2 mode the only accessible in set #2 is register 0x05 */
			if( v==5 )
				chip->address = v | 0x100;
			else
				chip->address = v;	/* verified range: 0x01, 0x04, 0x20-0xef(set #2 becomes set #1 in opl2 mode) */
		}
	break;
	}

	return chip->status>>7;
}

static unsigned char OPL3Read(OPL3lite *chip,int a)
{
	if( a==0 )
	{
		if (chip->st[0]) {
			if (chip->TC[0]) chip->TC[0]--;
			else {
				chip->TC[0]=chip->T[0]*20;
				OPL3_STATUS_SET(chip,0x40);
			}
		}
		if (chip->st[1]) {
			if (chip->TC[1]) chip->TC[1]--;
			else {
				chip->TC[1]=chip->T[1]*20;
				OPL3_STATUS_SET(chip,0x40);
			}
		}
		return chip->status;
	}

	return 0x00;	/* verified on real YMF262 */
}



static int OPL3TimerOver(OPL3lite *chip,int c)
{
	if( c )
	{	/* Timer B */
		OPL3_STATUS_SET(chip,0x20);
	}
	else
	{	/* Timer A */
		OPL3_STATUS_SET(chip,0x40);
	}
	/* reload timer */
	return chip->status>>7;
}


#if (BUILD_YMF262)

static OPL3lite *YMF262lite[MAX_OPL3_CHIPS];	/* array of pointers to the YMF262's */

 extern "C" {
 extern int Status940;
 extern int g_hMemory;
 extern volatile unsigned short *g_pusRegs;
 extern unsigned char *g_pSharedMemory ;
 extern void UpdateThreadEntry(void);
 extern  void Pause940(int n);
 extern  void Reset940(int yes);
 extern  void Startup940();
 extern  void Shutdown940();
 extern  void CleanUp(void);
 extern  void InitSharedMemory();
}

static void InitMemory()
{
	SharedBuff_ptr = (char *) (g_pSharedMemory + BUFF_BASE_ADDRESS);
	SharedData_ptr = (char *) (g_pSharedMemory + DATA_BASE_ADDRESS);
	memset(SharedBuff_ptr,0, END_OFFSET);
	memset(SharedData_ptr,0, END_OFFSET2);
}

int YMF262Init(int num, int clock, int rate)
{
	int i;
        char *ptr;

	if (YMF262NumChips)
		return -1;	/* duplicate init. */

        if(END_OFFSET>OPL3_MSG_SIZE ||
	   END_OFFSET2>OPL3_DAT_SIZE) {
	 LOG_MSG("OPL3 memory data error");
	 return -1;
	} 

        memset(counter,0,8);

        LOG_MSG("OPL3 reports");
        LOG_MSG("OPL3 mem: %d %d %d %d", sizeof(OPL3_SLOT),
	 sizeof(OPL3_CH),sizeof(OPL3),OPL3_SIZE);

        InitSharedMemory();
        InitMemory();
		
        ptr=(SharedData_ptr + NUMCHIP_OFFSET);
	YMF262NumChips=(int *) ptr;
	*YMF262NumChips = num;

        ptr=(SharedBuff_ptr + NSUB_OFFSET);
	NSubmittedMessages=(int *) ptr;
        *NSubmittedMessages=-1;
	
        ptr=(SharedBuff_ptr + NEX_OFFSET);
	NExecutedMessages=(int *) ptr;
	*NExecutedMessages=-1; 

        ptr=(SharedBuff_ptr + MSG_BUF_OFFSET);
	MessageBuffer=(CoreMessage *) ptr;

        ptr=(SharedBuff_ptr + TL_TAB_OFFSET);
	tl_tab=(signed int *) ptr;
        ptr=(SharedBuff_ptr + SIN_TAB_OFFSET);
        sin_tab=(unsigned int *) ptr;

        init_tables();

	for (i = 0;i < *YMF262NumChips; i++)
	{
	        ptr=(SharedBuff_ptr + BUFPOS_OFFSET+ i*sizeof(int));
		BufWritePos[i]=(int *) ptr;
		*BufWritePos[i]=0; 

	        ptr=(SharedBuff_ptr + READPOS_OFFSET +i*sizeof(int));
		BufReadPos[i]=(int *) ptr;
		*BufReadPos[i]=0; 

                ptr=(SharedBuff_ptr + DATA_OFFSET + 
		 i * SHARED_BUF_SIZE * sizeof(INT32)); 
		SharedBuffer[i]=(INT32 *) ptr;

                ptr=(SharedData_ptr + OPL_OFFSET + i*OPL3_SIZE);
		YMF262[i] = (OPL3*) ptr;
		
		ptr = (char *) malloc(sizeof(OPL3lite));
		memset(ptr , 0, sizeof(OPL3lite));
		YMF262lite[i] = (OPL3lite *) ptr;

	        ptr=(SharedBuff_ptr + FNTAB_OFFSET+i*1024*sizeof(UINT32));
	        YMF262lite[i]->fn_tab=(UINT32 *) ptr;

		OPL3Create(OPL3_TYPE_YMF262,clock,rate,YMF262lite[i],YMF262[i]);
        }
        LOG_MSG("dat: %d", YMF262[0]->eg_timer_overflow);

        LOG_MSG("OPL3:: %d %d", *YMF262NumChips,Status940);
	UpdateThreadEntry();

	ADD_MESSAGE(INIT ,num ,clock ,rate );

	for (i = 0;i < *YMF262NumChips; i++)
	{
		YMF262ResetChip(i);
	}

	return 0;
}

void YMF262Shutdown(void)
{
	int i;

        LOG_MSG("OPL3 ...");
	ADD_MESSAGE(SHUTDOWN,0,0,0);
	WAIT_FOR_SYNC;
	LOG_MSG("OPL3 end");

	for (i = 0;i < *YMF262NumChips; i++)
	{
		/* emulator shutdown */
		OPL3Destroy(YMF262lite[i]);
		YMF262[i] = NULL;
		YMF262lite[i] = NULL;
	}
	*YMF262NumChips = 0;
	CleanUp();
}

void YMF262ResetChip(int which)
{
	ADD_MESSAGE(RESET,which,0,0);
	OPL3ResetChip(YMF262lite[which]);
}

int YMF262Write(int which, int a, int v)
{
        ADD_MESSAGE(WRITE,which,a,v);
	return OPL3Write(YMF262lite[which], a, v);
}

unsigned char YMF262Read(int which, int a)
{
        ADD_MESSAGE(READ,which,a,0);
	return OPL3Read(YMF262lite[which], a);
}

int YMF262TimerOver(int which, int c)
{
        ADD_MESSAGE(TIMEROVER,which,c,0);
	return OPL3TimerOver(YMF262lite[which], c);
}

void YMF262SetTimerHandler(int which, OPL3_TIMERHANDLER TimerHandler, int channelOffset)
{
	OPL3SetTimerHandler(YMF262lite[which], TimerHandler, channelOffset);
}

void YMF262SetIRQHandler(int which,OPL3_IRQHANDLER IRQHandler,int param)
{
	OPL3SetIRQHandler(YMF262lite[which], IRQHandler, param);
}

void YMF262SetUpdateHandler(int which,OPL3_UPDATEHANDLER UpdateHandler,int param)
{
	OPL3SetUpdateHandler(YMF262lite[which], UpdateHandler, param);
}

void YMF262UpdateOne(int which, INT16 *buf16, int length)
{
 int i,ncopy,nfree;
 static int fillbufonly=1;
 static int nn=0;
 INT32 *buffer=(INT32 *) buf16;  /* stereo hence each sample is 2 buffer pos */
 static int warn=1;
 
 if(warn && *NSubmittedMessages-*NExecutedMessages>MSG_BUF_SIZE) 
 {
  LOG_MSG("OPL3: buffer running full"); 
  warn=0;
 } else
 {
  warn=1;
 } 
 ADD_MESSAGE(UPDATE,which, MIN(length,SHARED_BUF_SIZE),0);

 if(WRAPPED( *BufWritePos[which]- *BufReadPos[which],SHARED_BUF_SIZE) > SHARED_BUF_SIZE-3*length)
  fillbufonly=0;

 if(fillbufonly)
 {
  for(i=0;i<length;i++) buffer[i]=0;
  return; 
 }

 ncopy=MIN(length ,WRAPPED( *BufWritePos[which]-*BufReadPos[which],SHARED_BUF_SIZE));
 if(ncopy<length){
    LOG_MSG(" -1- %d %d",ncopy,length);
    fillbufonly=1;
 }
 nfree=SHARED_BUF_SIZE - *BufReadPos[which];
 if(ncopy < nfree) 
 {
   for(i=0;i<ncopy;i++) buffer[i]=SharedBuffer[which][ (*BufReadPos[which])+i ];
   (*BufReadPos[which])+=ncopy;
 } else
 {
   for(i=0;i<nfree;i++) buffer[i]=SharedBuffer[which][ (*BufReadPos[which])+i ];	      
   for(i=0;i<ncopy-nfree;i++) buffer[i+nfree]=SharedBuffer[which][i];
   *BufReadPos[which]=ncopy-nfree;
 }
 for(i=ncopy;i<length;i++) buffer[i]=buffer[i-1];
 
}


#endif /* BUILD_YMF262 */
